# Copyright 2005 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

import utils
import config
import xdb
if(utils.checkTwisted()):
	from twisted.words.protocols.jabber import jid
else:
	from tlib.jabber import jid

import shutil
import os
import os.path


def init():
	global noteList
	global noteListF

	try:
		notes = NotesToMyself()
		for note in noteList:
			if notes.check(note):
				noteListF[noteList.index(note)]()
				notes.append(note)
		notes.save()
	except:
		print "An error occurred during one of the automatic data update routines. Please report this bug."
		raise


class NotesToMyself:
	def __init__(self):
		pre = utils.doPath(config.spooldir) + "/" + config.jid + "/"
		self.filename = pre + "/notes_to_myself"
		self.notes = []
		
		if os.path.exists(self.filename):
			f = open(self.filename, "r")
			self.notes = [x.strip() for x in f.readlines()]
			f.close()
		elif not os.path.exists(pre):
			global noteList
			self.notes = noteList
			os.makedirs(pre)
	
	def check(self, note):
		return self.notes.count(note) == 0
	
	def append(self, note):
		if self.check(note):
			self.notes.append(note)

	def save(self):
		f = open(self.filename, "w")
		for note in self.notes:
			f.write(note + "\n")
		f.close()



def doSpoolPrepCheck():
	pre = utils.doPath(config.spooldir) + "/" + config.jid + "/"

	print "Checking spool files and stringprepping any if necessary...",

	for file in os.listdir(pre):
		file = xdb.unmangle(file).decode("utf-8")
		filej = jid.JID(file).full()
		if(file != filej):
			file = xdb.mangle(file)
			filej = xdb.mangle(filej)
			if(os.path.exists(filej)):
				print "Need to move", file, "to", filej, "but the latter exists!\nAborting!"
				os.exit(1)
			else:
				shutil.move(pre + file, pre + filej)
	print "done"


def doHashDirUpgrade():
	print "Upgrading your XDB structure to use hashed directories for speed...",

	# Do avatars...
	pre = utils.doPath(config.spooldir) + "/" + config.jid + "/avatars/"
	if os.path.exists(pre):
		for file in os.listdir(pre):
			if os.path.isfile(pre + file):
				pre2 = pre + file[0:3] + "/"
				if not os.path.exists(pre2):
					os.makedirs(pre2)
				shutil.move(pre + file, pre2 + file)
	
	# Do spool files...
	pre = utils.doPath(config.spooldir) + "/" + config.jid + "/"
	if os.path.exists(pre):
		for file in os.listdir(pre):
			if os.path.isfile(pre + file) and file != "notes_to_myself":
				hash = file[0:2]
				pre2 = pre + hash + "/"
				if not os.path.exists(pre2):
					os.makedirs(pre2)

				if(os.path.exists(pre2 + file)):
					print "Need to move", file, "to", pre2 + file, "but the latter exists!\nAborting!"
					os.exit(1)
				else:
					shutil.move(pre + file, pre2 + file)

	print "done"



noteList = ["doSpoolPrepCheck", "doHashDirUpgrade"]
noteListF = [doSpoolPrepCheck, doHashDirUpgrade]

