# Copyright 2004-2005 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

import utils
if utils.checkTwisted():
	from twisted.xish.domish import Element
else:
	from tlib.domish import Element
from tlib import msn
from legacy import glue
import avatar
import debug


# Default avatar for MSN contacts
f = open("legacy/defaultAvatar.png")
defaultAvatarData = f.read()
f.close()
defaultAvatar = avatar.AvatarCache().setAvatar(defaultAvatarData)


def getGroupNames(msnContact, msnContactList):
	groups = []
	for groupGUID in msnContact.groups:
		try:
			groups.append(msnContactList.groups[groupGUID])
		except KeyError:
			pass
	return groups

def msnlist2jabsub(lists):
	""" Converts MSN contact lists ORed together into the corresponding Jabber subscription state """
	if lists & msn.FORWARD_LIST and lists & msn.REVERSE_LIST:
		return "both"
	elif lists & msn.REVERSE_LIST:
		return "from"
	elif lists & msn.FORWARD_LIST:
		return "to"
	else:
		return "none"


def jabsub2msnlist(sub):
	""" Converts a Jabber subscription state into the corresponding MSN contact lists ORed together """
	if sub == "to":
		return msn.FORWARD_LIST
	elif sub == "from":
		return msn.REVERSE_LIST
	elif sub == "both":
		return (msn.FORWARD_LIST | msn.REVERSE_LIST)
	else:
		return 0



class LegacyList:
	def __init__(self, session):
		self.session = session
		self.subscriptionBuffer = []
	
	def removeMe(self):
		self.subscriptionBuffer = None
		self.session = None

	def addContact(self, jid):
		debug.log("LegacyList: Session \"%s\" addContact(\"%s\")" % (self.session.jabberID, jid))
		userHandle = glue.jid2msn(jid)
		self.session.legacycon.addContact(msn.FORWARD_LIST, userHandle)
		self.session.contactList.getContact(jid).contactGrantsAuth()
	
	def removeContact(self, jid):
		debug.log("LegacyList: Session \"%s\" removeContact(\"%s\")" % (self.session.jabberID, jid))
		jid = glue.jid2msn(jid)
		self.session.legacycon.remContact(msn.FORWARD_LIST, jid)
	
	
	def authContact(self, jid):
		debug.log("LegacyList: Session \"%s\" authContact(\"%s\")" % (self.session.jabberID, jid))
		jid = glue.jid2msn(jid)
		d = self.session.legacycon.remContact(msn.PENDING_LIST, jid)
		if d:
			self.session.legacycon.addContact(msn.REVERSE_LIST, jid)
		self.session.legacycon.remContact(msn.BLOCK_LIST, jid)
		self.session.legacycon.addContact(msn.ALLOW_LIST, jid)
	
	def deauthContact(self, jid):
		debug.log("LegacyList: Session \"%s\" deauthContact(\"%s\")" % (self.session.jabberID, jid))
		jid = glue.jid2msn(jid)
		self.session.legacycon.remContact(msn.ALLOW_LIST, jid)
		self.session.legacycon.addContact(msn.BLOCK_LIST, jid)



	def syncJabberLegacyLists(self):
		""" Synchronises the MSN contact list on server with the Jabber contact list """

		global defaultAvatar

		# We have to make an MSNContactList from the XDB data, then compare it with the one the server sent
		# Any subscription changes must be sent to the client, as well as changed in the XDB
		debug.log("LegacyList: Session \"%s\" starting syncJabberLegacyLists()" % (self.session.jabberID))
		result = self.session.pytrans.xdb.request(self.session.jabberID, "jabber:iq:roster")
		oldContactList = msn.MSNContactList()
		if result:
			for item in result.elements():
				user = item.getAttribute("jid")
				sub = item.getAttribute("subscription")
				lists = item.getAttribute("lists")
				if not lists:
					lists = jabsub2msnlist(sub) # Backwards compatible
				lists = int(lists)
				contact = msn.MSNContact(userHandle=user, screenName="", lists=lists)
				oldContactList.addContact(contact)
		
		newXDB = Element((None, "query"))
		newXDB.attributes["xmlns"] = "jabber:iq:roster"
		
		contactList = self.session.legacycon.getContacts()


		# Convienence functions
		def addedToList(num):
			return (not (oldLists & num) and (lists & num))
		def removedFromList(num):
			return ((oldLists & num) and not (lists & num))
		
		for contact in contactList.contacts.values():
			# Compare with the XDB <item/> entry
			oldContact = oldContactList.getContact(contact.userHandle)
			if oldContact == None:
				oldLists = 0
			else:
				oldLists = oldContact.lists
			lists = contact.lists
			
			# Create the Jabber representation of the
			# contact base on the old list data and then
			# sync it with current
			jabContact = self.session.contactList.createContact(glue.msn2jid(contact.userHandle), msnlist2jabsub(oldLists))
			jabContact.updateAvatar(defaultAvatar, push=False)

			if addedToList(msn.FORWARD_LIST):
				jabContact.syncGroups(getGroupNames(contact, contactList), push=False)
				jabContact.syncContactGrantedAuth()

			if removedFromList(msn.FORWARD_LIST):
				jabContact.syncContactRemovedAuth()

			if addedToList(msn.ALLOW_LIST):
				jabContact.syncUserGrantedAuth()

			if addedToList(msn.BLOCK_LIST) or removedFromList(msn.ALLOW_LIST):
				jabContact.syncUserRemovedAuth()

			if (not (lists & msn.ALLOW_LIST) and not (lists & msn.BLOCK_LIST) and (lists & msn.REVERSE_LIST)) or (lists & msn.PENDING_LIST):
				jabContact.contactRequestsAuth()

			if removedFromList(msn.REVERSE_LIST):
				jabContact.contactDerequestsAuth()
			
			item = newXDB.addElement("item")
			item.attributes["jid"] = contact.userHandle
			item.attributes["subscription"] = msnlist2jabsub(lists)
			item.attributes["lists"] = str(lists)
		
		# Update the XDB
		self.session.pytrans.xdb.set(self.session.jabberID, "jabber:iq:roster", newXDB)
		debug.log("LegacyList: Session \"%s\" finished syncJabberLegacyLists()" % (self.session.jabberID))
	
	def saveLegacyList(self):
		contactList = self.session.legacycon.getContacts()
		if not contactList: return

		newXDB = Element((None, "query"))
		newXDB.attributes["xmlns"] = "jabber:iq:roster"
	
		for contact in contactList.contacts.values():
			item = newXDB.addElement("item")
			item.attributes["jid"] = contact.userHandle
			item.attributes["subscription"] = msnlist2jabsub(contact.lists) # Backwards compat
			item.attributes["lists"] = str(contact.lists)

		self.session.pytrans.xdb.set(self.session.jabberID, "jabber:iq:roster", newXDB)
		debug.log("LegacyList: Session \"%s\" finished saving list.")
	

