# Copyright 2004-2005 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

from twisted.internet import reactor
from twisted.internet.protocol import ClientFactory
from tlib import msn
import math
import base64
import binascii
import math
import config
import utils
import lang
import debug

MAXMESSAGESIZE = 1400


class MSNConnection:
	""" Manages all the Twisted factories, etc """
	def __init__(self, username, password):
		self.username = username
		self.password = password
		self.inited = False
		self.tries = 0
		self.initMe()
		debug.log("MSNConnection: \"%s\" created" % (self.username))
		
	def initMe(self):
		if self.inited:
			MSNConnection.removeMe(self)
		
		self.switchboardSessions = {}
		dispatchFactory = DispatchFactory(self)
		reactor.connectTCP('messenger.hotmail.com', 1863, dispatchFactory)
		self.notificationFactory = msn.NotificationFactory()
		self.notificationFactory.userHandle = self.username
		self.notificationFactory.password = self.password
		self.notificationFactory.msncon = self
		self.notificationFactory.protocol = Notification
		self.notificationFactory.initialListVersion = self.initialListVersion
		self.notificationProtocol = None
		
		self.savedEvents = SavedEvents()
		
		self.inited = True
		
		debug.log("MSNConnection: \"%s\" initialised" % (self.username))
	
	def removeMe(self):
		debug.log("MSNConnection: \"%s\" destroyed" % (self.username))
		if self.notificationProtocol:
			self.notificationProtocol.removeMe()
		if self.notificationFactory:
			self.notificationFactory.msncon = None
		self.notificationFactory = None
		self.notificationProtocol = None
		self.savedEvents = SavedEvents()
		for userHandle in self.switchboardSessions.copy():
			self.switchboardSessions[userHandle].removeMe()
		self.switchboardSessions = {}
	
	def resourceOffline(self, offlineResource):
		for contact in self.switchboardSessions.keys():
			if self.switchboardSessions[contact].resource == offlineResource:
				self.switchboardSessions[contact].resource = self.highestResource()
	
	def getContacts(self):
		if self.notificationFactory:
			return self.notificationFactory.contacts
		else:
			return None
		
	
	def sendMessage(self, remoteUser, resource, text, noerror):
		debug.log("MSNConnection: \"%s\" sendMessage(\"%s\", \"%s\")" % (self.username, remoteUser, text))
		if self.notificationProtocol:
			if not self.switchboardSessions.has_key(remoteUser):
				self.switchboardSessions[remoteUser] = SwitchboardSession(self, remoteUser, resource)
			self.switchboardSessions[remoteUser].resource = resource
			self.switchboardSessions[remoteUser].sendMessage(text.replace("\n", "\r\n"), noerror)
		elif not noerror:
			self.failedMessage(remoteUser, text)
	
	def requestAvatar(self, userHandle):
		debug.log("MSNConnection: \"%s\" requestAvatar(\"%s\")" % (self.username, userHandle))
		resource = self.session.highestResource()
		if config.getAllAvatars:
			if not self.switchboardSessions.has_key(userHandle):
				self.switchboardSessions[userHandle] = SwitchboardSession(self, userHandle, resource)
			else:
				self.switchboardSessions[userHandle].requestAvatar()
		else:
			if self.switchboardSessions.has_key(userHandle): # Only request avatars for open switchboards
				self.switchboardSessions[userHandle].requestAvatar()
	
	def sendTypingToContact(self, remoteUser):
		if self.switchboardSessions.has_key(remoteUser):
			self.switchboardSessions[remoteUser].sendTypingNofication()
	
	def notificationProtocolReady(self, notificationProtocol):
		self.notificationProtocol = notificationProtocol
		self.loggedIn()
		self.tries = 0
	
	def sendSavedEvents(self):
		# Hack for events sent before we're logged in
		self.savedEvents.send(self)
		self.savedEvents = None
	
	def changeAvatar(self, imageData):
		if self.notificationProtocol:
			self.notificationProtocol.changeAvatar(imageData, push=True)
		else:
			self.savedEvents.avatarImageData = imageData
	
	def changeStatus(self, statusCode, screenName, personal):
		if self.notificationProtocol:
			def cb1(arg):
				self.ourStatusChanged(arg[0])
			def cb2(arg):
				self.ourNickChanged(arg[0])
			def cb3(arg):
				self.ourPersonalChanged(personal)
			debug.log("MSNConnection: \"%s\" - changing status and screenName (\"%s\", \"%s\", \"%s\")" % (self.username, statusCode, screenName, personal))
			if statusCode:
				statusCode = str(statusCode.encode("utf-8"))
				self.notificationProtocol.changeStatus(statusCode).addCallback(cb1)
			if screenName:
				screenName = str(screenName.encode("utf-8"))
				self.notificationProtocol.changeScreenName(screenName).addCallback(cb2)
			if personal:
				personal = str(personal.encode("utf-8"))
			else:
				personal = ""
			self.notificationProtocol.changePersonalMessage(personal).addCallback(cb3)
		else:
			self.savedEvents.statusCode = statusCode
			self.savedEvents.screenName = screenName
			self.savedEvents.personal = personal
	
	def connectionLostBase(self, reason):
		# Attempts to reconnect
		if self.tries < 5 and self.session:
			reactor.callLater(2 ** self.tries, self.initMe)
			self.tries += 1
		else:
			self.connectionLost(self)
	
	def addContact(self, listType, userHandle):
		if self.notificationProtocol:
			return self.notificationProtocol.addContact(listType, str(userHandle))
		else:
			self.savedEvents.addContacts.append((listType, str(userHandle)))
	
	def remContact(self, listType, userHandle, groupID=0):
		if self.notificationProtocol:
			return self.notificationProtocol.remContact(listType, str(userHandle))
		else:
			self.savedEvents.remContacts.append((listType, str(userHandle)))
	
	
	
	def initialEmailNotification(self, inboxunread, foldersunread):
		pass
	
	def realtimeEmailNotification(self, mailfrom, fromaddr, subject):
		pass
	
	def loggedIn(self):
		pass
	
	def loginFailure(self, message):
		pass
		
	def multipleLogin(self):
		pass
	
	def gotMessage(self, remoteUser, resource, text):
		pass
	
	def avatarHashChanged(self, userHandle, hash):
		pass
	
	def gotAvatarImage(self, to, image):
		pass
	
	def listSynchronized(self):
		pass
	
	def contactStatusChanged(self, remoteUser):
		pass
	
	def ourStatusChanged(self, statusCode):
		pass
	
	def ourNickChanged(self, nick):
		pass
	
	def ourPersonalChanged(self, personal):
		pass
	
	def userMapping(self, passport, jid):
		pass
	
	def gotContactTyping(self, remoteUser, resource):
		pass
	
	def serverGoingDown(self):
		pass
	
	def accountNotVerified(self):
		pass
	
	def userAddedMe(self, userHandle):
		pass
	
	def userRemovedMe(self, userHandle):
		pass
	
	def failedMessage(self, remoteUser, message):
		pass
	
	def connectionLost(self):
		pass



class SavedEvents:
	def __init__(self):
		self.nickname = ""
		self.statusCode = ""
		self.personal = ""
		self.avatarImageData = ""
		self.addContacts = []
		self.remContacts = []
	
	def send(self, msncon):
		if self.avatarImageData:
			msncon.notificationProtocol.changeAvatar(self.avatarImageData, push=False)
		if self.nickname or self.statusCode or self.personal:
			msncon.changeStatus(self.statusCode, self.nickname, self.personal)
		for listType, userHandle in self.addContacts:
			msncon.addContact(listType, userHandle)
		for listType, userHandle in self.remContacts:
			msncon.remContact(listType, userHandle)
			


def switchToGroupchat(switchboardSession, user1, user2):
	gcsbs = GroupchatSwitchboardSession()
	from glue import LegacyGroupchat, msn2jid
	groupchat = LegacyGroupchat(session=switchboardSession.msncon.session, resource=None, existing=True, switchboardSession=gcsbs)
	gcsbs.groupchat = groupchat
	gcsbs.msncon = switchboardSession.msncon
	gcsbs.switchboard = switchboardSession.switchboard
	gcsbs.switchboard.switchboardSession = gcsbs
	gcsbs.ready = True
	gcsbs.userJoined(user1)
	gcsbs.userJoined(user2)
	groupchat.sendUserInvite(msn2jid(switchboardSession.remoteUser))
	switchboardSession.removeMe(False)
	debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" created by conversion" % (gcsbs.groupchat.roomJID(), gcsbs))
	return gcsbs


class SwitchboardSessionBase:
	def sendMessage(self, message, noerror):
		if self.ready:
			def failedMessage(ignored):
				if self.__class__ == GroupchatSwitchboardSession:
					tempmsncon.failedMessage(self.groupchat.roomJID(), message)
				else:
					tempmsncon.failedMessage(self.remoteUser, message)
					
			tempmsncon = self.msncon # In case MSN tells us the message failed after removeMe()

			debug.log("SwitchboardSessionBase: \"%s\" sending message \"%s\"" % (self.msncon.username, message))
			message = str(message.encode("utf-8"))
			
			if len(message) < MAXMESSAGESIZE:
				msnmessage = msn.MSNMessage(message=message)
				msnmessage.setHeader("Content-Type", "text/plain; charset=UTF-8")
				msnmessage.ack = msn.MSNMessage.MESSAGE_NACK

				d = self.switchboard.sendMessage(msnmessage)
				if not noerror:
					d.addCallback(failedMessage)
			else:
				chunks = int(math.ceil(len(message) / float(MAXMESSAGESIZE)))
				chunk = 0
				guid = utils.random_guid()
				while chunk < chunks:
					offset = chunk * MAXMESSAGESIZE
					text = message[offset : offset + MAXMESSAGESIZE]

					msnmessage = msn.MSNMessage(message=text)
					msnmessage.setHeader("Message-ID", guid)
					if chunk == 0:
						msnmessage.setHeader("Content-Type", "text/plain; charset=UTF-8")
						msnmessage.setHeader("Chunks", str(chunks))
					else:
						msnmessage.setHeader("Chunk", str(chunk))
					msnmessage.ack = msn.MSNMessage.MESSAGE_NACK

					d = self.switchboard.sendMessage(msnmessage)
					if not noerror:
						d.addCallback(failedMessage)
					chunk += 1

			self.resetTimer()
		else:
			self.messageBuffer.append((message, noerror))
	
	def gotAvatarImage(self, to, image):
		self.msncon.gotAvatarImage(to, image)
	
	def resetTimer(self):
		pass


class GroupchatSwitchboardSession(SwitchboardSessionBase):
	def __init__(self, groupchat=None, makeSwitchboard=False):
		self.removed = False

		self.msncon = None
		self.groupchat = None
		if groupchat:
			self.groupchat = groupchat
			self.msncon = self.groupchat.session.legacycon
		self.switchboard = None
		self.ready = False
		self.messageBuffer = []
		self.invitedUsers = []
		self.oneUserHasJoined = False
		
		if makeSwitchboard and groupchat:
			debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" requesting a switchboard session" % (self.groupchat.roomJID(), self))
			d = self.msncon.notificationProtocol.requestSwitchboardServer()
			d.addCallback(self.sbRequestAccepted)
			d.addErrback(self.removeMe)
		
		if self.msncon:
			debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" created" % (self.msncon.username, self))
	
	def removeMe(self):
		if self.removed:
			debug.log("GroupchatSwitchboardSession: removeMe called more than once! Traceback!")
			return
		self.removed = True

		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" destroyed" % (self.groupchat.roomJID(), self))
		self.msncon = None
		if self.switchboard:
			self.switchboard.removeMe()
		self.switchboard = None
		self.groupchat = None
		self.ready = False

		utils.mutilateMe(self)
	
	def sbRequestAccepted(self, (host, port, key)):
		# Connect to the switchboard server
		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" sbRequestAccepted()" % (self.msncon.username, self))
		reactor.connectTCP(host, port, SwitchboardFactory(self, key))
	
	def sendMessage(self, message, noerror):
		if self.oneUserHasJoined:
			SwitchboardSessionBase.sendMessage(self, message, noerror)
		else:
			self.messageBuffer.append((message, noerror))
	
	def inviteUser(self, userHandle):
		userHandle = str(userHandle)
		if self.ready:
			debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" inviting %s" % (self.msncon.username, self, userHandle))
			self.switchboard.inviteUser(userHandle)
		else:
			self.invitedUsers.append(userHandle)
	
	def gotMessage(self, message):
		self.groupchat.messageReceived(message.userHandle, message.getMessage())
	
	def flushBuffer(self):
		for m, noerror in utils.copyList(self.messageBuffer):
			self.messageBuffer.remove((m, noerror))
			self.sendMessage(m, noerror)
		
		for i in utils.copyList(self.invitedUsers):
			self.invitedUsers.remove(i)
			self.inviteUser(i)
	
	def userJoined(self, userHandle):
		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" userJoined(\"%s\")" % (self.msncon.username, self, userHandle))
		self.oneUserHasJoined = True
		self.flushBuffer()
		self.groupchat.contactJoined(userHandle)
	
	def userLeft(self, userHandle):
		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" userLeft(\"%s\")" % (self.msncon.username, self, userHandle))
		self.groupchat.contactLeft(userHandle)



class SwitchboardSession(SwitchboardSessionBase):
	def __init__(self, msncon, remoteUser, resource, reply=False, host=None, port=None, key=None, sessionID=None):
		self.removed = False

		self.msncon = msncon
		self.remoteUser = str(remoteUser)
		self.resource = str(resource)
		
		self.killTimer = reactor.callLater(30.0*60.0, self.removeMe)
		
		self.switchboard = None # The SwitchboardClient class
		self.messageBuffer = [] # Any messages sent before the switchboard is ready are buffered
		self.ready = False # Is True when we are connected to the switchboard, and the remote user has accepted our invite
		
		if not reply:
			# Request a switchboard
			d = self.msncon.notificationProtocol.requestSwitchboardServer()
			d.addCallback(self.sbRequestAccepted)
			d.addErrback(self.removeMe)
		else:
			reactor.connectTCP(host, port, SwitchboardFactory(self, key, sessionID, reply))
		
		debug.log("SwitchboardSession: \"%s\" \"%s\" \"%s\" created" % (self.msncon.username, self.remoteUser, self.resource))
	
	def removeMe(self, sbflag=True):
		if self.removed:
			debug.log("SwitchboardSession: removeMe called more than once! Traceback!")
			return
		self.removed = True

		debug.log("SwitchboardSession: \"%s\" \"%s\" \"%s\" destroyed" % (self.msncon.username, self.remoteUser, self.resource))
		for message, noerror in self.messageBuffer:
			if not noerror:
				self.msncon.failedMessage(self.remoteUser, message)
		self.messageBuffer = []

		del self.msncon.switchboardSessions[self.remoteUser]
		self.msncon = None
		if sbflag and self.switchboard:
			self.switchboard.removeMe()
		self.switchboard = None
		self.ready = False
		if self.killTimer and not self.killTimer.called:
			self.killTimer.cancel()
		self.killTimer = None

		utils.mutilateMe(self)
	
	def resetTimer(self):
		# Sets a count down timer to kill this switchboard session in 30 minutes
		self.killTimer.cancel()
		self.killTimer = reactor.callLater(30.0*60.0, self.removeMe)
	
	def sbRequestAccepted(self, (host, port, key)):
		# Connect to the switchboard server
		reactor.connectTCP(host, port, SwitchboardFactory(self, key))
	
	def sendTypingNofication(self):
		if self.ready:
			self.switchboard.sendTypingNotification()
	
	def contactTyping(self):
		self.msncon.gotContactTyping(self.remoteUser, self.resource)
	
	def flushBuffer(self):
		for m, noerror in utils.copyList(self.messageBuffer):
			self.messageBuffer.remove((m, noerror))
			self.sendMessage(m, noerror)
	
	def gotMessage(self, message):
		self.msncon.gotMessage(self.remoteUser, self.resource, message.getMessage())
		self.resetTimer()
	
	CAPS = msn.MSNContact.MSNC1 | msn.MSNContact.MSNC2 | msn.MSNContact.MSNC3 | msn.MSNContact.MSNC4
	def requestAvatar(self):
		if not self.switchboard: return
		msnContacts = self.msncon.getContacts()
		if not msnContacts: return
		msnContact = msnContacts.getContact(self.remoteUser)
		if not (msnContact and msnContact.caps & self.CAPS and msnContact.msnobj): return
		if msnContact.msnobjGot: return
		msnContact.msnobjGot = True # This is deliberately set before we get the avatar. So that we don't try to reget failed avatars over & over
		self.switchboard.sendAvatarRequest(self.remoteUser, msnContact.msnobj)
	
	def userJoined(self, userHandle):
		if userHandle != self.remoteUser:
			# Another user has joined, so we now have three participants (these two and ourself)
			switchToGroupchat(self, self.remoteUser, userHandle)
		else:
			self.requestAvatar()
	
	def userLeft(self, userHandle):
		if userHandle == self.remoteUser:
			self.removeMe()



class DispatchFactory(ClientFactory):
	def __init__(self, msncon):
		self.msncon = msncon
	
	def buildProtocol(self, addr):
		p = Dispatch(self.msncon)
		del self.msncon # No longer needed
		return p
	
	def clientConnectionFailed(self, connector, reason):
		self.msncon.connectionLostBase(reason)
	

class Dispatch(msn.DispatchClient):
	def __init__(self, msncon):
		msn.DispatchClient.__init__(self)
		self.msncon = msncon
		self.userHandle = self.msncon.username
	
	def __del__(self):
		self.factory = None
		self.msncon = None
	
	def gotNotificationReferral(self, host, port):
		self.transport.loseConnection()
		if self.msncon and self.msncon.session and self.msncon.session.alive:
			reactor.connectTCP(host, port, self.msncon.notificationFactory)



class Notification(msn.NotificationClient):
	def __init__(self):
		self.removed = False

		msn.NotificationClient.__init__(self, proxy=config.proxyServer, proxyport=config.proxyPort)

	def removeMe(self):
		if self.removed:
			debug.log("Notification: removeMe called more than once! Traceback!")
			return
		self.removed = True

		self.logOut()
		self.transport.loseConnection()
		if self.factory.msncon:
			self.factory.msncon.notificationProtocol = None
			self.factory.msncon = None
		self.factory = None

		utils.mutilateMe(self)
	
	def badConditions(self):
		if not (self.factory and self.factory.msncon and self.factory.msncon.session and self.factory.msncon.session.alive):
			if not self.removed:
				self.removeMe()
			return True
		return False
		
	
	def loginFailure(self, message):
		if self.badConditions(): return
		self.factory.msncon.loginFailure(message)
	
	def loggedIn(self, userHandle, verified):
		if self.badConditions(): return

		self.factory.msncon.notificationProtocolReady(self)
		if not verified:
			self.factory.msncon.accountNotVerified()
		
		msn.NotificationClient.loggedIn(self, userHandle, verified)
		
		debug.log("NotificationClient: \"%s\" authenticated with MSN servers" % (self.factory.msncon.username))
	
	def msnAlertReceived(self, body, action, subscr):
		if self.badConditions(): return
		self.factory.msncon.msnAlert(body, action, subscr)

	def initialEmailNotification(self, inboxunread, foldersunread):
		if self.badConditions() or not config.mailNotifications: return
		self.factory.msncon.initialEmailNotification(inboxunread, foldersunread)

	def realtimeEmailNotification(self, mailfrom, fromaddr, subject):
		if self.badConditions() or not config.mailNotifications: return
		self.factory.msncon.realtimeEmailNotification(mailfrom, fromaddr, subject)
	
	def connectionLost(self, reason):
		if self.badConditions(): return
		def wait():
			debug.log("NotificationClient: \"%s\" lost connection with MSN servers" % (self.factory.userHandle))
			msn.NotificationClient.connectionLost(self, reason)
			self.factory.msncon.connectionLostBase(reason)
		# Make sure this event is handled after any others
		reactor.callLater(0, wait)
	
	def listSynchronized(self, *args):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" MSN contact lists synchronised" % (self.factory.userHandle))
		self.factory.msncon.listSynchronized()
		if self.badConditions(): return # Just in case the session is deregistered
		self.factory.msncon.sendSavedEvents()
		self.setPrivacyMode(False)
	
	def gotSwitchboardInvitation(self, sessionID, host, port, key, remoteUser, screenName):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" gotSwitchboardInvitation(\"%s\")" % (self.factory.userHandle, remoteUser))
		sbs = SwitchboardSession(self.factory.msncon, remoteUser, self.factory.msncon.session.highestResource(), True, host, port, key, sessionID)
		if self.factory.msncon.switchboardSessions.has_key(remoteUser):
			self.factory.msncon.switchboardSessions[remoteUser].removeMe()
		self.factory.msncon.switchboardSessions[remoteUser] = sbs
	
	def avatarHashChanged(self, userHandle, hash):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" avatarHashChanged(\"%s\")" % (self.factory.userHandle, userHandle))
		hash = base64.decodestring(hash)
		hash = binascii.hexlify(hash)
		self.factory.msncon.avatarHashChanged(userHandle, hash)
	
	def contactStatusChanged(self, statusCode, userHandle, screenName):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" contactStatusChanged(\"%s\", \"%s\")" % (self.factory.userHandle, statusCode, userHandle))
		
		self.factory.msncon.contactStatusChanged(userHandle)
	
	def contactPersonalChanged(self, userHandle, personal):
		if self.badConditions(): return
		msn.NotificationClient.contactPersonalChanged(self, userHandle, personal)
		self.factory.msncon.contactStatusChanged(userHandle)
	
	def contactOffline(self, userHandle):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" contactOffline(\"%s\")" % (self.factory.userHandle, userHandle))
		msn.NotificationClient.contactOffline(self, userHandle)
		
		self.factory.msncon.contactStatusChanged(userHandle)
	
	def userAddedMe(self, userGuid, userHandle, screenName):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" userAddedMe(\"%s\", \"%s\")" % (self.factory.userHandle, userGuid, userHandle))
		msn.NotificationClient.userAddedMe(self, userGuid, userHandle, screenName)
		self.factory.msncon.userAddedMe(userHandle)
	
	def userRemovedMe(self, userGuid, userHandle):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" userRemovedMe(\"%s\")" % (self.factory.userHandle, userHandle))
		msn.NotificationClient.userRemovedMe(self, userGuid, userHandle)
		self.factory.msncon.userRemovedMe(userHandle)
	
	def multipleLogin(self):
		if self.badConditions(): return
		debug.log("NotificationClient: \"%s\" multiple logins" % (self.factory.msncon.username))
		self.factory.msncon.multipleLogin()



class SwitchboardFactory(ClientFactory):
	def __init__(self, switchboardSession, key, sessionID=None, reply=False):
		self.switchboardSession = switchboardSession
		self.key = key
		self.sessionID = sessionID
		self.reply = reply
	
	def buildProtocol(self, addr):
		p = Switchboard(self.switchboardSession)
		if p.badConditions(): return p
		p.key = self.key
		p.sessionID = self.sessionID
		p.reply = self.reply
		p.userHandle = self.switchboardSession.msncon.username
		p.factory = self
		return p

class Switchboard(msn.SwitchboardClient):
	def __init__(self, switchboardSession):
		self.removed = False

		self.switchboardSession = switchboardSession
		self.chattingUsers = []
		self.callid = None
		msn.SwitchboardClient.__init__(self)
		if self.badConditions(): return
		self.msnobj = self.switchboardSession.msncon.notificationProtocol.msnobj
		debug.log("SwitchboardClient: \"%s\" \"%s\" - created" % (self.switchboardSession.msncon.username, self.switchboardSession))
	
	def removeMe(self):
		if self.removed:
			debug.log("Switchboard: removeMe called more than once! Traceback!")
			return
		self.removed = True

		self.transport.loseConnection()
		debug.log("SwitchboardClient: \"%s\" - destroyed" % (self.switchboardSession))
		self.switchboardSession = None
		self.factory.switchboardSession = None
		self.factory = None

		if self.callid and not self.callid.called:
			self.callid.cancel() # Cancel the invite fail message
		self.callid = None

		utils.mutilateMe(self)
	
	def badConditions(self):
		if not (self.switchboardSession and self.switchboardSession.msncon and self.switchboardSession.msncon.session and self.switchboardSession.msncon.session.alive):
			if self.switchboardSession:
				if not self.switchboardSession.removed:
					self.switchboardSession.removeMe()
			elif not self.removed:
				self.removeMe()
			return True
		return False
	
	def loggedIn(self):
		if self.badConditions(): return
		if (not self.reply) and self.switchboardSession.__class__ == SwitchboardSession:
			def failCB(arg=None):
				debug.log(templogmessage)
				self.switchboardSession.removeMe()
			d = self.inviteUser(self.switchboardSession.remoteUser)
			d.addErrback(failCB)
			templogmessage = "SwitchboardClient: \"%s\" \"%s\" - user has NOT joined after 30 seconds" % (self.switchboardSession.msncon.username, self.switchboardSession.remoteUser)
			# If the user doesn't join then we want to tear down the SwitchboardSession
			self.callid = reactor.callLater(30.0, failCB)
		
		else:
			self.readySwitchboardSession()
	
	def readySwitchboardSession(self, ignored=None):
		if self.badConditions(): return
		debug.log("SwitchboardClient: \"%s\" \"%s\" - ready for use" % (self.switchboardSession.msncon.username, self.switchboardSession))
		self.switchboardSession.ready = True
		self.switchboardSession.switchboard = self
		self.switchboardSession.flushBuffer()
		for user in self.chattingUsers:
			self.switchboardSession.userJoined(user)
		if self.callid and not self.callid.called:
			self.callid.cancel() # Cancel the invite fail message (only applies if we needed to invite the user)
		self.callid = None
	
	def gotChattingUsers(self, users):
		for user in users:
			self.chattingUsers.append(user)
	
	def userJoined(self, userHandle, screenName):
		if self.badConditions(): return
		if (not self.reply) and self.switchboardSession.__class__ == SwitchboardSession:
			self.readySwitchboardSession()
		debug.log("SwitchboardClient: \"%s\" \"%s\" - userJoined(\"%s\")" % (self.switchboardSession.msncon.username, self.switchboardSession, userHandle))
		self.switchboardSession.userJoined(userHandle)
	
	def userLeft(self, userHandle):
		if self.badConditions(): return
		debug.log("SwitchboardClient: \"%s\" \"%s\" - userLeft(\"%s\")" % (self.switchboardSession.msncon.username, self.switchboardSession, userHandle))
		def wait():
			if self.badConditions(): return
			self.switchboardSession.userLeft(userHandle)
		# Make sure this event is handled after any others (eg, gotMessage)
		reactor.callLater(0, wait)
	
	def gotMessage(self, message):
		if self.badConditions():
			debug.log("SwitchboardClient: gotMessage called too late! Traceback!")
			return
		debug.log("SwitchboardClient: \"%s\" \"%s\" gotMessage(\"%s\")" % (self.switchboardSession.msncon.username, message.userHandle, message.getMessage()))
		cTypes = [s.lstrip() for s in message.getHeader("Content-Type").split(';')]
		if "text/plain" in cTypes:
			try:
				if len(cTypes) > 1 and cTypes[1].find("UTF-8") >= 0:
					message.message = message.message.decode("utf-8")
				self.switchboardSession.gotMessage(message)
			except:
				self.switchboardSession.gotMessage(lang.get(self.switchboardSession.msncon.session.lang).msnDroppedMessage) # FIXME, this is a little deep
				raise
			return
		if "text/x-clientcaps" in cTypes:
			if message.hasHeader("JabberID"):
				jid = message.getHeader("JabberID")
				self.switchboardSession.msncon.userMapping(message.userHandle, jid)
			return
		debug.log("Discarding unknown message type: %s" % (message.getMessage()))
	
	def userTyping(self, message):
		if self.badConditions(): return
		if self.switchboardSession.__class__ == SwitchboardSession: # Ignore typing in groupchats
			if message.userHandle == self.switchboardSession.remoteUser:
				self.switchboardSession.contactTyping()
	
	def sendClientCaps(self):
		message = msn.MSNMessage()
		message.setHeader("Content-Type", "text/x-clientcaps")
		message.setHeader("Client-Name", "PyMSNt")
		message.setHeader("JabberID", str(self.switchboardSession.msncon.session.jabberID)) # FIXME, this is a little deep
		self.sendMessage(message)
	
	def sendMessage(self, message):
		# A little bit of fancyness to make sure that clientcaps
		# only gets sent after the first text message.
		if message.getHeader("Content-Type").startswith("text"):
			self.sendMessage = type(self.sendMessage)(msn.SwitchboardClient.sendMessage, self, Switchboard)
			self.sendClientCaps()
			return self.sendMessage(message)
		else:
			return msn.SwitchboardClient.sendMessage(self, message)
	
	def gotAvatarImage(self, to, image):
		if self.badConditions(): return
		self.switchboardSession.gotAvatarImage(to, image)

 
