# Copyright 2004-2005 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

import utils
import os
import shutil
import time
if os.name == "posix":
	import signal
import sys
reload(sys)
sys.setdefaultencoding("utf-8")

# Must load config before everything else
import config
import xmlconfig
xmlconfig.reloadConfig()

if config.reactor == "epoll":
	from twisted.internet import epollreactor
	epollreactor.install()
elif config.reactor == "poll":
	from twisted.internet import pollreactor
	pollreactor.install()
elif config.reactor == "kqueue":
	from twisted.internet import kqreactor
	kqreactor.install()
elif len(config.reactor) > 0:
	print "Unknown reactor: ", config.reactor, ". Using default, select(), reactor."

from twisted.internet import reactor, task
from twisted.internet.defer import Deferred
import twisted.python.log
if utils.checkTwisted():
	from twisted.words.protocols.jabber import component, jid
	from twisted.xish.domish import Element
else:
	from tlib.jabber import component, jid
	from tlib.domish import Element


import xdb
import avatar
import session
import jabw
import disco
import register
import misciq
import lang
import debug
import legacy
import housekeep

#import gc
#gc.set_debug(gc.DEBUG_COLLECTABLE | gc.DEBUG_UNCOLLECTABLE | gc.DEBUG_INSTANCES | gc.DEBUG_OBJECTS)


class PyTransport(component.Service):
	def __init__(self):
		debug.log("PyTransport: Service starting up")
		
		# Discovery, as well as some builtin features
		self.discovery = disco.ServerDiscovery(self)
		self.discovery.addIdentity("gateway", legacy.id, legacy.name, config.jid)
		self.discovery.addIdentity("conference", "text", legacy.name + " Chatrooms", config.jid)
		self.discovery.addFeature("jabber:x:conference", None, config.jid) # So that clients know you can create groupchat rooms on the server
		self.discovery.addFeature("jabber:iq:conference", None, config.jid) # So that clients know you can create groupchat rooms on the server
		
		self.xdb = xdb.XDB(config.jid, legacy.mangle)
		self.avatarCache = avatar.AvatarCache()
		self.registermanager = register.RegisterManager(self)
		self.gatewayTranslator = misciq.GatewayTranslator(self)
		self.versionTeller = misciq.VersionTeller(self)
		self.pingService = misciq.PingService(self)
		self.adHocCommands = misciq.AdHocCommands(self)
		self.vCardFactory = misciq.VCardFactory(self)
		self.iqAvatarFactor = misciq.IqAvatarFactory(self)
		self.connectUsers = misciq.ConnectUsers(self)
		self.statistics = misciq.Statistics(self)
		self.startTime = int(time.time())

		self.xmlstream = None
		self.sessions = {}
		
		# Groupchat ID handling
		self.lastID = 0
		self.reservedIDs = []

		# Message IDs
		self.messageID = 0
		
		self.loopCall = task.LoopingCall(self.loopCall)
		self.loopCall.start(60.0)

		if config.debugOn:
			twisted.python.log.addObserver(self.exceptionLogger)
		
	
	def removeMe(self):
		debug.log("PyTransport: Service shutting down")
		for session in self.sessions.copy():
			self.sessions[session].removeMe()
	
	def exceptionLogger(self, *kwargs):
		if len(config.debugLog) > 0:
			kwargs = kwargs[0]
			if kwargs.has_key("failure"):
				failure = kwargs["failure"]
				failure.printTraceback(debug) # Pass debug as a pretend file object because it implements the write method
				if config.debugLog:
					debug.flushDebugSmart()
					print "Exception occured! Check the log!"
	
	def makeMessageID(self):
		self.messageID += 1
		return str(self.messageID)
	
	def makeID(self):
		newID = "r" + str(self.lastID)
		self.lastID += 1
		if self.reservedIDs.count(newID) > 0:
			# Ack, it's already used.. Try again
			return self.makeID()
		else:
			return newID
	
	def reserveID(self, ID):
		self.reservedIDs.append(ID)
	
	def loopCall(self):
		numsessions = len(self.sessions)

		if config.debugOn and numsessions > 0:
			debug.log("Sessions:")
			for key in self.sessions:
				debug.log("\t" + self.sessions[key].jabberID)
	
		self.statistics.stats["Uptime"] = int(time.time()) - self.startTime
		self.statistics.stats["OnlineUsers"] = numsessions
		legacy.updateStats(self.statistics)
		if numsessions > 0:
			oldDict = self.sessions.copy()
			self.sessions = {}
			for key in oldDict:
				session = oldDict[key]
				if not session.alive:
					debug.log("Ghost session %s found. This shouldn't happen. Trace" % (session.jabberID))
					# Don't add it to the new dictionary. Effectively removing it
				else:
					self.sessions[key] = session
	
	def componentConnected(self, xmlstream):
		debug.log("PyTransport: Connected to main Jabberd server")
		self.xmlstream = xmlstream
		self.xmlstream.addObserver("/iq", self.discovery.onIq)
		self.xmlstream.addObserver("/presence", self.onPresence)
		self.xmlstream.addObserver("/message", self.onMessage)
		self.xmlstream.addObserver("/route", self.onRouteMessage)
		if config.useXCP:
			pres = Element((None, "presence"))
			pres.attributes["to"] = "presence@-internal"
			pres.attributes["from"] = config.compjid
			x = pres.addElement("x")
			x.attributes["xmlns"] = "http://www.jabber.com/schemas/component-presence.xsd"
			x.attributes["xmlns:config"] = "http://www.jabber.com/config"
			x.attributes["config:version"] = "1"
			x.attributes["protocol-version"] = "1.0"
			x.attributes["config-ns"] = legacy.url + "/component"
			self.send(pres)
	
	def componentDisconnected(self):
		debug.log("PyTransport: Disconnected from main Jabberd server")
		self.xmlstream = None
	
	def onRouteMessage(self, el):
		for child in el.elements():
			if child.name == "message":
				self.onMessage(child)
			elif child.name == "presence":
				# Ignore any presence broadcasts about other XCP components
				if child.getAttribute("to") and child.getAttribute("to").find("@-internal") > 0: return
				self.onPresence(child)
			elif child.name == "iq":
				self.discovery.onIq(child)
	
	def onMessage(self, el):
		fro = el.getAttribute("from")
		try:
			froj = jid.JID(fro)
		except Exception, e:
			debug.log("PyTransport: Failed stringprep on <message from=\"%s\"/> - %s" % (fro, str(e)))
			return
		mtype = el.getAttribute("type")
		if self.sessions.has_key(froj.userhost()):
			self.sessions[froj.userhost()].onMessage(el)
		elif mtype != "error":
			to = el.getAttribute("to")
			ulang = utils.getLang(el)
			body = None
			for child in el.elements():
				if child.name == "body":
					body = child.__str__()
			debug.log("PyTrans: Sending error response to a message outside of session.")
			jabw.sendErrorMessage(self, fro, to, "auth", "not-authorized", lang.get(ulang).notLoggedIn, body)
	
	def onPresence(self, el):
		fro = el.getAttribute("from")
		to = el.getAttribute("to")
		try:
			froj = jid.JID(fro)
			toj = jid.JID(to)
		except Exception, e:
			debug.log("PyTransport: Failed stringprep on <presence from=\"%s\" to=\"%s\"/> - %s" % (fro, to, str(e)))
			return

		if self.sessions.has_key(froj.userhost()):
			self.sessions[froj.userhost()].onPresence(el)
		else:
			ulang = utils.getLang(el)
			ptype = el.getAttribute("type")
			if to.find('@') < 0:
				# If the presence packet is to the transport (not a user) and there isn't already a session
				if not el.getAttribute("type"): # Don't create a session unless they're sending available presence
					debug.log("PyTransport: Attempting to create a new session \"%s\"" % (froj.userhost()))
					s = session.makeSession(self, froj.userhost(), ulang)
					if s:
						self.statistics.stats["TotalUsers"] += 1
						self.sessions[froj.userhost()] = s
						debug.log("PyTransport: New session created \"%s\"" % (froj.userhost()))
						# Send the first presence
						s.onPresence(el)
					else:
						debug.log("PyTransport: Failed to create session \"%s\"" % (froj.userhost()))
						jabw.sendMessage(self, to=froj.userhost(), fro=config.jid, body=lang.get(ulang).notRegistered)
				
				elif el.getAttribute("type") != "error":
					debug.log("PyTransport: Sending unavailable presence to non-logged in user \"%s\"" % (froj.userhost()))
					pres = Element((None, "presence"))
					pres.attributes["from"] = to
					pres.attributes["to"] = fro
					pres.attributes["type"] = "unavailable"
					self.send(pres)
					return
			
			elif ptype and (ptype.startswith("subscribe") or ptype.startswith("unsubscribe")):
				# They haven't logged in, and are trying to change subscription to a user
				# Lets log them in and then do it
				debug.log("PyTransport: Attempting to create a session to do subscription stuff %s" % (froj.userhost()))
				s = session.makeSession(self, froj.userhost(), ulang)
				if s:
					self.sessions[froj.userhost()] = s
					debug.log("PyTransport: New session created \"%s\"" % (froj.userhost()))
					# Tell the session there's a new resource
					s.handleResourcePresence(froj.userhost(), froj.resource, toj.userhost(), toj.resource, 0, None, None, None)
					# Send this subscription
					s.onPresence(el)


class App:
	def __init__(self):
		# Check that there isn't already a PID file
		if os.path.isfile(utils.doPath(config.pid)):
			pf = open(utils.doPath(config.pid))
			pid = int(str(pf.readline().strip()))
			pf.close()
			if os.name == "posix":
				try:
					os.kill(pid, signal.SIGHUP)
					self.alreadyRunning()
				except OSError:
					# The process is still up
					pass
			else:
				self.alreadyRunning()
		
		# Create a PID file
		pid = str(os.getpid())
		pf = file(utils.doPath(config.pid), 'w')
		pf.write("%s\n" % pid);
		pf.close()
	
		jid = config.jid
		if config.compjid: jid = config.compjid
		self.c = component.buildServiceManager(jid, config.secret, "tcp:%s:%s" % (config.mainServer, config.port))
		self.transportSvc = PyTransport()
		self.transportSvc.setServiceParent(self.c)
		self.c.startService()
		reactor.addSystemEventTrigger('before', 'shutdown', self.shuttingDown)
	
	def alreadyRunning(self):
		print "There is already a transport instance running with this configuration."
		print "Exiting..."
		sys.exit(1)
	
	def shuttingDown(self):
		self.transportSvc.removeMe()
		def cb(ignored=None):
			os.remove(utils.doPath(config.pid))
		d = Deferred()
		d.addCallback(cb)
		reactor.callLater(3.0, d.callback, None)
		return d



def SIGHUPstuff(*args):
	xmlconfig.reloadConfig()
	debug.reopenFile()


if __name__ == "__main__":
	# Set SIGHUP to reload the config file & close & open debug file
	if os.name == "posix":
		signal.signal(signal.SIGHUP, SIGHUPstuff)
	
	# Do any auto-update stuff
	housekeep.init()

	app = App()
	reactor.run()


